from itertools import zip_longest
from typing import Collection, Optional, Iterable, Any, Dict, Callable, Union

from sqlalchemy.orm import load_only, Load
from sqlalchemy import column, desc, and_, or_, not_, Select, inspect

from assimilator.alchemy.database.model_utils import get_model_from_relationship
from assimilator.alchemy.database.specifications.filtering_options import AlchemyFilteringOptions
from assimilator.core.database.specifications.specifications import (
    specification,
    SpecificationList,
    SpecificationType,
    FilterSpecification,
)


class AlchemyFilter(FilterSpecification):
    filtering_options_cls = AlchemyFilteringOptions

    def __init__(self, *filters, **named_filters):
        super(AlchemyFilter, self).__init__(*filters)
        self.filters.append(named_filters)

    def __or__(self, other: 'FilterSpecification') -> SpecificationType:
        return CompositeFilter(self, other, func=or_)

    def __and__(self, other: 'FilterSpecification') -> SpecificationType:
        return CompositeFilter(self, other, func=and_)

    def __invert__(self):
        return CompositeFilter(self, func=not_)

    def parse_filters(self, model):
        self.filtering_options.table_name = str(inspect(model).selectable)
        named_filters = list(filter_ for filter_ in self.filters if isinstance(filter_, Dict))

        for filter_ in named_filters:
            for field, value in filter_.items():
                self.filters.append(
                    self.filtering_options.parse_field(
                        raw_field=field, value=value,
                    )
                )

            self.filters.remove(filter_)

    def __call__(self, query: Select, **context: Any) -> Select:
        self.parse_filters(context['repository'].model)
        return query.filter(*self.filters)


class CompositeFilter(AlchemyFilter):
    def __init__(self, *filters: Union[FilterSpecification, 'CompositeFilter'], func: Callable):
        super(CompositeFilter, self).__init__()
        self.filter_specs = filters
        self.func = func
    
    def __call__(self, query: Select, **context: Any) -> Select:
        parsed_specs = []

        for spec in self.filter_specs:
            parsed_specs.append(spec(query=Select(), **context).whereclause)

        return query.filter(self.func(*parsed_specs))


alchemy_filter = AlchemyFilter


@specification
def alchemy_order(*clauses: str, query: Select, **_) -> Select:
    parsed_clauses = []

    for clause in clauses:
        if clause.startswith("-"):
            parsed_clauses.append(desc(column(clause[1:], is_literal=True)))
        else:
            parsed_clauses.append(column(clause, is_literal=True))

    return query.order_by(*parsed_clauses)


@specification
def alchemy_paginate(
    *,
    limit: Optional[int] = None,
    offset: Optional[int] = None,
    query: Select,
    **_,
) -> Select:
    if offset is not None:
        query = query.offset(offset)
    if limit is not None:
        query = query.limit(limit)

    return query


@specification
def alchemy_join(
    *targets: Collection,
    join_args: Iterable[dict] = None,
    query: Select,
    **context,
) -> Select:
    model = context['repository'].model

    for target, join_data in zip_longest(targets, (join_args or {}), fillvalue=dict()):
        if not target:
            continue

        if isinstance(target, str):
            entities = target.split(".")
            target = model

            for entity in entities:
                target, _ = get_model_from_relationship(
                    model=target,
                    relationship_name=entity,
                )

        query = query.join_from(model, target, **join_data)\
            .add_columns(target).select_from(model)

    return query


@specification
def alchemy_only(
    *only_fields: str,
    query: Select,
    model,
    **_,
):
    models_to_fields: Dict[Load, Any] = {}
    parsed_loads = list(field for field in only_fields if not isinstance(field, str))

    if parsed_loads:
        query = query.options(load_only(*parsed_loads))

    for field in (field for field in only_fields if isinstance(field, str)):
        parts = field.split('.')
        if len(parts) == 1:
            models_to_fields[Load(model)] = [
                getattr(model, field),
                *models_to_fields.get(model, []),
            ]
            continue

        current_load = model
        for part in parts:
            model_relationships = set(inspect(current_load).relationships.keys())

            if part in model_relationships:
                current_load, _ = get_model_from_relationship(current_load, relationship_name=part)
            else:
                models_to_fields[Load(current_load)] = [
                    getattr(current_load, part),
                    *models_to_fields.get(current_load, []),
                ]
                break

        else:
            models_to_fields[Load(current_load)] = []

    for loader, load_fields in models_to_fields.items():
        if load_fields:
            query = query.options(loader.load_only(*load_fields))
        else:
            query = query.options(loader)

    return query


class AlchemySpecificationList(SpecificationList):
    filter = AlchemyFilter
    order = alchemy_order
    paginate = alchemy_paginate
    join = alchemy_join
    only = alchemy_only


__all__ = [
    'AlchemySpecificationList',
    'alchemy_filter',
    'AlchemyFilter',
    'alchemy_order',
    'alchemy_paginate',
    'alchemy_join',
    'alchemy_only',
]



import operator
from typing import Optional, Iterable, Union, Callable, Any

from assimilator.core.database.specifications.specifications import specification, FilterSpecification


class AdaptiveFilter:
    def __init__(self, *fields, **kwargs_fields):
        self.fields = fields
        self.kwargs_fields = kwargs_fields

    def __or__(self, other: Union['AdaptiveFilter', 'FilterSpecification']) -> 'CompositeAdaptiveFilter':
        return CompositeAdaptiveFilter(first=self, second=other, func=operator.or_)

    def __and__(self, other: Union['AdaptiveFilter', 'FilterSpecification']) -> 'CompositeAdaptiveFilter':
        return CompositeAdaptiveFilter(first=self, second=other, func=operator.and_)

    def __invert__(self):
        return AdaptiveFilter(self.fields, self.kwargs_fields)

    def __call__(self, query, repository, **context):
        return repository.specs.filter(
            *self.fields, **self.kwargs_fields,
        )(query=query, repository=repository)


class CompositeAdaptiveFilter(AdaptiveFilter):
    def __init__(
        self,
        first: Union['AdaptiveFilter', 'FilterSpecification'],
        second: Union['AdaptiveFilter', 'FilterSpecification'],
        func: Callable[['AdaptiveFilter', 'AdaptiveFilter'], Any],
    ):
        super(CompositeAdaptiveFilter, self).__init__()
        self.first = first
        self.second = second
        self.func = func

    def _parse_specification(self, filter_spec, repository):
        if isinstance(filter_spec, CompositeAdaptiveFilter):
            first = self._parse_specification(filter_spec=filter_spec.first, repository=repository)
            second = self._parse_specification(filter_spec=filter_spec.second, repository=repository)
            return filter_spec.func(first, second)

        elif isinstance(filter_spec, AdaptiveFilter):
            return repository.specs.filter(*filter_spec.fields, **filter_spec.kwargs_fields)
        else:
            return filter_spec

    def __call__(self, query, repository, **context):
        first = self._parse_specification(filter_spec=self.first, repository=repository)
        second = self._parse_specification(filter_spec=self.first, repository=repository)
        return self.func(first, second)(query=query, repository=repository, **context)


filter_ = AdaptiveFilter


@specification
def order(*clauses: str, query, repository, **context):
    return repository.specs.order(*clauses)(query=query, repository=repository, **context)


@specification
def paginate(
    *,
    limit: Optional[int] = None,
    offset: Optional[int] = None,
    query,
    repository,
    **context,
):
    paginate_spec = repository.specs.paginate(limit=limit, offset=offset)
    return paginate_spec(query=query, repository=repository, **context)


@specification
def join(*targets: str, join_args: Iterable[dict] = None, query, repository, **context):
    return repository.specs.join(*targets, join_args=join_args)(query=query, repository=repository, **context)


@specification
def only(*only_fields: str, query, repository, **context):
    return repository.specs.only(*only_fields)(query=query, repository=repository, **context)


__all__ = [
    'AdaptiveFilter',
    'filter_',
    'only',
    'order',
    'join',
    'paginate',
]



from typing import ClassVar

from bson import ObjectId
from pydantic import Field

from assimilator.core.database.models import BaseModel


class MongoModel(BaseModel):
    upsert: bool = False
    id: ObjectId = Field(allow_mutation=False)

    class AssimilatorConfig:
        collection: ClassVar[str]
        autogenerate_id: ClassVar[bool] = True
        exclude = {'collection', 'upsert'}

    def __hash__(self):
        return int(str(self.id), base=16)

    def generate_id(self, **kwargs) -> ObjectId:
        return ObjectId()


__all__ = ['MongoModel']



from operator import or_, and_
from typing import Union, List, Generator, Any

from assimilator.core.database.models import BaseModel
from assimilator.core.database import FilterSpecification
from assimilator.internal.database.specifications.internal_operator import invert
from assimilator.internal.database.specifications.filtering_options import InternalFilteringOptions

QueryT = Union[str, List[BaseModel]]


class InternalFilter(FilterSpecification):
    filtering_options_cls = InternalFilteringOptions

    def __init__(self, *filters, **named_filters):
        self.text_filters = [filter_ for filter_ in filters if isinstance(filter_, str)]

        if named_filters.get('id'):
            self.text_filters.append(named_filters.pop('id'))

        super(InternalFilter, self).__init__(
            *(set(filters) - set(self.text_filters)),
            **named_filters,
        )

    def __call__(self, query: QueryT, **context) -> Union[str, Generator[BaseModel, Any, None]]:
        if isinstance(query, str):
            return f'{query}{"".join(self.text_filters)}'
        elif not self.filters:
            return query

        return (
            model for model in query
            if all(filter_func(model) for filter_func in self.filters)
        )

    def __or__(self, other: Union['InternalFilter', 'CompositeFilter']) -> 'InternalFilter':
        return CompositeFilter(first=self, second=other, operation=or_)

    def __and__(self, other: Union['InternalFilter', 'CompositeFilter']) -> 'InternalFilter':
        return CompositeFilter(first=self, second=other, operation=and_)

    def __invert__(self):
        return InternalFilter(*(invert(func) for func in self.filters))


class CompositeFilter(InternalFilter):
    def __init__(
        self,
        first: Union[FilterSpecification, 'CompositeFilter'],
        second: Union[FilterSpecification, 'CompositeFilter'],
        operation: Union[or_, and_],
    ):
        super(CompositeFilter, self).__init__()
        self.first = first
        self.second = second
        self.operation = operation

    def __call__(self, query: QueryT, **context) -> Union[str, QueryT]:
        if isinstance(query, str):
            first_result = self.first(query=query, **context)
            second_result = self.second(query=query, **context)
            return f'{query}{first_result.replace(query, "")}{second_result.replace(query, "")}'

        first_result = self.first(query=query, **context)
        second_result = self.second(query=query, **context)

        return list(self.operation(set(first_result), set(second_result)))

    def __str__(self):
        return f"{self.first} {self.operation} {self.second}"


__all__ = ['InternalFilter']



import sqlalchemy

if sqlalchemy.__version__ < "2.0.0":
    raise RuntimeWarning(
        "PyAssimilator will only support SQLAlchemy 2 from now on. Please, update "
        "the library using this manual: https://docs.sqlalchemy.org/en/20/changelog/migration_20.html"
    )



from typing import Dict, List, Set, Tuple, Iterable

from assimilator.core.database.models import BaseModel


InternalContainers = (List, Set, Tuple, map)


def find_model_value(fields: Iterable[str], model: BaseModel):
    model_val = model

    for foreign_field in fields:
        if isinstance(model_val, InternalContainers):
            model_val = list(getattr(obj, foreign_field) for obj in model_val)
        elif isinstance(model_val, Dict):
            model_val = list(getattr(obj, foreign_field) for obj in model_val.values())
        else:
            model_val = getattr(model_val, foreign_field)

    return model_val



import os


if os.environ.get('PY_ASSIMILATOR_MESSAGE', None) is None:
    print(f"""
                 ___              _           _ __      __            
    ____  __  __/   |  __________(_)___ ___  (_) /___ _/ /_____  _____
   / __ \/ / / / /| | / ___/ ___/ / __ `__ \/ / / __ `/ __/ __ \/ ___/
  / /_/ / /_/ / ___ |(__  |__  ) / / / / / / / / /_/ / /_/ /_/ / /    
 / .___/\__, /_/  |_/____/____/_/_/ /_/ /_/_/_/\__,_/\__/\____/_/     
/_/    /____/


    Thank you for using PyAssimilator!
    Documentation is available here: https://knucklesuganda.github.io/py_assimilator/
    Star this library on GitHub and get perks: https://github.com/knucklesuganda/py_assimilator

    If you want to turn off this text, add an environment variable:
    1) You can add this code BEFORE all of your imports:
        import os
        os.environ['PY_ASSIMILATOR_MESSAGE'] = 'False'

    2) You can add it within your system or in your .env file: PY_ASSIMILATOR_MESSAGE=False
    
    """)



import operator
import re
from functools import wraps
from numbers import Number
from typing import Any, Callable, Union, Literal

from assimilator.core.database.models import BaseModel
from assimilator.core.database.specifications.filtering_options import FILTERING_OPTIONS_SEPARATOR
from assimilator.internal.database.specifications.utils import InternalContainers, find_model_value


def find_attribute(func: callable, field: str, value: Any) -> Callable[[BaseModel], bool]:
    """
    That decorator is used to get the value of the field from a BaseModel that is provided in the query.
    We do that because we need to get the value of the field in the internal specifications, not just the name
    of it. For example, User(id=1) will use field='id' to get 1 as the result.

    :param func: filtering option function that is going to be decorated.
    :param field: field name that is used in getattr(model, field)
    :param value: value of the field.
    :return: function to be called with a model to find an attribute and call the comparison function.
    """

    @wraps(func)
    def find_attribute_wrapper(model: BaseModel) -> bool:
        foreign_fields = field.split(FILTERING_OPTIONS_SEPARATOR)

        if len(foreign_fields) == 1:
            return func(getattr(model, foreign_fields[0]), value)

        model_val = find_model_value(fields=foreign_fields, model=model)
        if isinstance(model_val, InternalContainers):
            return any(func(member, value) for member in model_val)

        return func(model_val, value)

    find_attribute_wrapper: func
    return find_attribute_wrapper


def eq(field: str, value: Any):
    return find_attribute(func=operator.eq, field=field, value=value)


def gt(field: str, value: Number):
    return find_attribute(func=operator.gt, field=field, value=value)


def gte(field: str, value: Number):
    return find_attribute(func=operator.ge, field=field, value=value)


def lt(field: str, value: Number):
    return find_attribute(func=operator.lt, field=field, value=value)


def lte(field: str, value: Number):
    return find_attribute(func=operator.le, field=field, value=value)


def not_(field: str, value: Union[Literal[True], Literal[False], Literal[None]]):
    return find_attribute(func=operator.not_, field=field, value=value)


def is_(field: str, value: Union[Literal[True], Literal[False], Literal[None]]):
    return find_attribute(func=operator.is_, field=field, value=value)


def regex(field: str, value: str):
    return find_attribute(
        func=lambda model_val, val: re.compile(value).match(model_val),
        field=field,
        value=value,
    )


def like(field: str, value: str):
    return regex(field, f'^{value.replace("%", ".*?")}$')


def invert(func: Callable):

    @wraps(func)
    def invert_wrapper(model):
        return not func(model)

    invert_wrapper: func
    return invert_wrapper


__all__ = [
    'find_attribute',
    'eq',
    'gt',
    'gte',
    'lt',
    'lte',
    'not_',
    'is_',
    'regex',
    'like',
    'invert',
]



from typing import Optional

from assimilator.core.patterns import ErrorWrapper
from assimilator.core.database.unit_of_work import UnitOfWork
from assimilator.redis_.database.repository import RedisRepository
from assimilator.internal.database.error_wrapper import InternalErrorWrapper


class RedisUnitOfWork(UnitOfWork):
    repository: RedisRepository

    def __init__(
        self,
        repository: RedisRepository,
        error_wrapper: Optional[ErrorWrapper] = None,
        autocommit: bool = False,
    ):
        super(RedisUnitOfWork, self).__init__(
            repository=repository,
            error_wrapper=error_wrapper or InternalErrorWrapper(),
            autocommit=autocommit,
        )

    def begin(self):
        self.repository.transaction = self.repository.session.pipeline()

    def rollback(self):
        self.repository.transaction.discard()

    def commit(self):
        self.repository.transaction.execute()

    def close(self):
        self.repository.transaction.reset()
        self.repository.transaction = self.repository.session


__all__ = [
    'RedisUnitOfWork',
]



from typing import Optional

from pymongo.client_session import ClientSession

from assimilator.core.database import UnitOfWork
from assimilator.core.patterns import ErrorWrapper
from assimilator.mongo.database.repository import MongoRepository
from assimilator.mongo.database.error_wrapper import MongoErrorWrapper


class MongoUnitOfWork(UnitOfWork):
    repository: MongoRepository
    transaction: ClientSession

    def __init__(
        self,
        repository: MongoRepository,
        error_wrapper: Optional[ErrorWrapper] = None,
        autocommit: bool = False,
    ):
        super(MongoUnitOfWork, self).__init__(
            repository=repository,
            error_wrapper=error_wrapper or MongoErrorWrapper(),
            autocommit=autocommit,
        )

    def begin(self):
        self.transaction = self.repository.session.start_session()
        self.transaction.start_transaction()

    def rollback(self):
        self.transaction.abort_transaction()

    def commit(self):
        self.transaction.commit_transaction()

    def close(self):
        pass


__all__ = [
    'MongoUnitOfWork',
]



import json
from uuid import uuid4, UUID
from typing import (
    Type, TypeVar, ClassVar, Union,
    Optional, Callable, Any, AbstractSet,
    Mapping, Dict,
)

from pydantic import BaseModel as PydanticBaseModel, Extra, ValidationError, Field

from assimilator.core.exceptions import ParsingError


T = TypeVar("T", bound='BaseModel')
AbstractSetIntStr = AbstractSet[Union[int, str]]
MappingIntStrAny = Mapping[Union[int, str], Any]


class BaseModel(PydanticBaseModel):
    id: str = Field(allow_mutation=False)

    class AssimilatorConfig(PydanticBaseModel, extra=Extra.allow):
        autogenerate_id: ClassVar[bool] = True
        exclude: ClassVar[set] = None

    class Config:
        arbitrary_types_allowed = True
        validate_assignment = True

    def __hash__(self):
        return UUID(self.id).int

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

        if not issubclass(cls.AssimilatorConfig, BaseModel.AssimilatorConfig):
            base_configs = [
                getattr(base_class, 'AssimilatorConfig') for base_class in cls.mro()
                if hasattr(base_class, 'AssimilatorConfig')
            ]

            class InheritedConfig(*base_configs):
                ...

            cls.AssimilatorConfig = InheritedConfig

        return cls

    def generate_id(self, **kwargs) -> str:
        return str(uuid4())

    def __init__(self, **kwargs):
        if self.AssimilatorConfig.autogenerate_id and kwargs.get('id') is None:
            kwargs['id'] = self.generate_id(**kwargs)

        super(BaseModel, self).__init__(**kwargs)

    @classmethod
    def loads(cls: Type['T'], data: str) -> 'T':
        try:
            return cls(**json.loads(data))
        except (ValidationError, TypeError) as exc:
            raise ParsingError(exc)

    def json(
        self,
        *,
        include: Optional[Union['AbstractSetIntStr', 'MappingIntStrAny']] = None,
        exclude: Optional[Union['AbstractSetIntStr', 'MappingIntStrAny']] = None,
        by_alias: bool = False,
        skip_defaults: Optional[bool] = None,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        encoder: Optional[Callable[[Any], Any]] = None,
        models_as_dict: bool = True,
        **dumps_kwargs: Any,
    ) -> str:
        return super(BaseModel, self).json(
            include=include,
            exclude={*(exclude or []), *(self.AssimilatorConfig.exclude or [])},
            by_alias=by_alias,
            skip_defaults=skip_defaults,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            encoder=encoder,
            models_as_dict=models_as_dict,
            **dumps_kwargs,
        )

    def dict(
        self,
        *,
        include: Optional[Union['AbstractSetIntStr', 'MappingIntStrAny']] = None,
        exclude: Optional[Union['AbstractSetIntStr', 'MappingIntStrAny']] = None,
        by_alias: bool = False,
        skip_defaults: Optional[bool] = None,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
    ) -> Dict[str, Any]:
        return super(BaseModel, self).dict(
            include=include,
            exclude={*(exclude or []), *(self.AssimilatorConfig.exclude or [])},
            by_alias=by_alias,
            skip_defaults=skip_defaults,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
        )


__all__ = [
    'BaseModel',
]
